﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using Stride.Core.Mathematics;
using VL.Lib.Mathematics;


namespace ForceDirected
{
    public class FDGraph
    {
        private List<FDNode> FNodes;
        private List<FDEdge> FEdges;

        private Octree FTree;

        private Random FRand;

        public List<FDNode> Nodes
        {
            get { return FNodes; }
        }

        public List<FDEdge> Edges
        {
            get { return FEdges; }
        }

        public Octree Tree
        {
            get { return FTree; }
        }

        public float Speed { get; set; }
        public string RootID { get; set; }


        public FDGraph()
        {
            FNodes = new List<FDNode>();
            FEdges = new List<FDEdge>();

            FRand = new Random();
        }


        public void SetForces(AttractionForce attraction, RepulsionForce repulsion)
        {
            Forces.Attraction = attraction;
            Forces.Repulsion = repulsion;
        }


        public void SetEades(Vector3 eades)
        {
            Forces.ConstantsEades = eades;
        }


        public void SetFruchtermanReingold(Vector2 fruchtermanReingold)
        {
            Forces.ConstantsFR = fruchtermanReingold;
        }


        public void SetSpringAttractor(Vector4 springAttractor)
        {
            Forces.ConstantsSA = springAttractor;
        }


        public void Clear()
        {
            FEdges.Clear();
            FNodes.Clear();
        }


        public void AddEdges(IReadOnlyList<string> f, IReadOnlyList<string> t)
        {
            var fromCount = f.Count;
            var toCount = t.Count;
            var edgeCount = Math.Max(fromCount, toCount);

            for (int i = 0; i < edgeCount; i++)
            {
                AddEdge(f[i%fromCount], t[i%toCount]);
            }
        }


        //add an edge by two ids
        public void AddEdge(string f, string t)
        {
            if (f == t) return;
            if (EdgeExists(f, t)) return;

            var fn = GetNode(f);
            var tn = GetNode(t);

            if (fn == null)
            {
                fn = new FDNode(f, IDToPos(f), 1.0f);
                FNodes.Add(fn);
            }

            if (tn == null)
            {
                tn = new FDNode(t, IDToPos(t), 1.0f);
                FNodes.Add(tn);
            }

            FEdges.Add(new FDEdge(fn, tn));
        }

        public bool EdgeExists(string f, string t)
        {
            var edgeCount = FEdges.Count;
            for (int i = 0; i < edgeCount; i++)
            {
                var e = FEdges[i];
                if ((e.From.ID == f && e.To.ID == t) ||
                   (e.From.ID == t && e.To.ID == f)) return true;

            }
            return false;
        }


        //get node by id
        public FDNode GetNode(string id)
        {
            var nodeCount = FNodes.Count;
            for (int i = 0; i < nodeCount; i++)
            {
                if (FNodes[i].ID == id) return FNodes[i];
            }
            return null;
        }


        protected Vector3 IDToPos(string id)
        {
            var pos = new Vector3(0f);

            if (id == RootID) return pos;

            pos = RandomUtil.NextVector3(FRand, new Vector3(-5), new Vector3(5));

            return pos;
        }


        public void RemoveNode(FDNode n)
        {
            foreach (var e in n.Parent)
            {
                e.From.Children.Remove(e);
                FEdges.Remove(e);
            }
            foreach (var e in n.Children)
            {
                e.To.Parent.Remove(e);
                FEdges.Remove(e);
            }
            FNodes.Remove(n);
        }


        public void UpdateOctree(float deltaTime, float theta = 0.8f)
        {

            Vector3 min = new Vector3(float.MaxValue);
            Vector3 max = new Vector3(float.MinValue);
            // min/max pos of all nodes to determine size of octree root boundary
            Parallel.ForEach(FNodes, node =>
            {
                Vector3 pos = node.Position;

                Vector3.Min(ref min, ref pos, out min);
                Vector3.Max(ref max, ref pos, out max);

            });

            // recreate octree every frame
            FTree = new Octree(min, max, 0, theta);

            FTree.BuildTree(FNodes);

            //attraction along edges
            Parallel.ForEach(FEdges, edge =>
            {
                edge.Update();
            });

            //repulsion of the nodes
            Parallel.ForEach(FNodes, node =>
            {                                               // seems to be necessary
                node.Velocity += FTree.CalcRepulsion(node) / FNodes.Count;
                node.Velocity *= Speed;

                //update position
                node.Update(deltaTime);
            });

        }


        public void UpdateBruteForce(float deltaTime)
        {
            //attraction along edges
            Parallel.ForEach(FEdges, edge =>
            {
                edge.Update();
            });

            //repulsion of the nodes
            Parallel.ForEach(FNodes, node =>
            {
                node.Velocity += CalcRepulsion(node);
                node.Velocity *= Speed;

                //update position
                node.Update(deltaTime);
            });
        }


        private Vector3 CalcRepulsion(FDNode n)
        {
            var rep = new Vector3(0);

            if (n.Parent.Count == 0) return rep;

            var oCount = FNodes.Count;
            for (int i = 0; i < oCount; i++)
            {
                var o = FNodes[i];

                if (o != n)
                {
                    var diff = n.Position - o.Position;
                    rep += Forces.Repulse(diff);
                }
            }

            return rep / oCount;
        }

    }
}
